import 'package:firebase_auth/firebase_auth.dart';
import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/cupertino.dart';
import 'package:nui_firebase/nui_firebase.dart';
import 'package:nui_firebase/nui_firebase_models.dart';

typedef FireAuthStateChangeListener = Function(bool signedIn, FireUser user);

class NUIFireAuth implements NUIFireComponent{

  NUIFireAuth._builder(NUIFireAuthBuilder builder);

  static NUIFireAuth _authInstance;

  static NUIFireAuth get(){
    if(_authInstance == null) throw Exception("NUIFireAuth -> NUIFireAuth has not been initialized yet. Call NUIFireAuth.Builder to initialize");
    return _authInstance;
  }

  FirebaseAuth instance(){
    return FirebaseAuth.instance;
  }

  @override
  String name(){
    return "NUIFireAuth";
  }

  FireUser currentUser(){
    if(instance().currentUser != null){
      return FireUser(instance().currentUser);
    }
    return null;
  }

  bool signedIn(){
    return currentUser() != null;
  }

  void listen(FireAuthStateChangeListener listener){
    if(listener != null) {
      instance()
          .authStateChanges()
          .listen((User user) {
        listener(user != null, user != null ? FireUser(user) : null);
      });
    }
  }

  Future<FireResult<UserCredential>> signInAnonymously() async{
    UserCredential user = await instance().signInAnonymously();
    if(user != null){
      return FireResult<UserCredential>(success: true, message: "Sign in anonymously successful", data: user);
    }
    else{
      return FireResult<UserCredential>(success: false, message: "Failed to sign in anonymously", data: null);
    }
  }

  Future<FireResult<UserCredential>> signUp({String email, String password}) async{
    FireResult<UserCredential> result;
    try {
      UserCredential userCredential = await instance().createUserWithEmailAndPassword(
          email: email,
          password: password
      );
      result = FireResult(success: true, message: "Sign up successful", data: userCredential);
    } on FirebaseAuthException catch (e) {
      if (e.code == 'weak-password') {
        result = FireResult(success: false, message: "Please provide a stronger password", data: null);
      } else if (e.code == 'email-already-in-use') {
        result = FireResult(success: false, message: "This email is already in use", data: null);
      } else {
        result = FireResult(success: false, message: e.toString(), data: null);
      }
    } catch (e) {
      result = FireResult(success: false, message: e.toString(), data: null);
    }

    return result;
  }

  Future<FireResult<UserCredential>> signIn({String email, String password}) async{
    FireResult<UserCredential> result;
    try {
      UserCredential userCredential = await instance().signInWithEmailAndPassword(
          email: email,
          password: password
      );
      result = FireResult(success: true, message: "Sign in successful", data: userCredential);
    } on FirebaseAuthException catch (e) {
      if (e.code == 'user-not-found') {
        result = FireResult(success: false, message: "No user found with this email", data: null);
      } else if (e.code == 'wrong-password') {
        result = FireResult(success: false, message: "Incorrect password was entered", data: null);
      } else {
        result = FireResult(success: false, message: e.toString(), data: null);
      }
    } catch(e){
      result = FireResult(success: false, message: e.toString(), data: null);
    }
    return result;
  }

  Future<FireResult> sendPasswordReset({@required String email}) async{
    FireResult result;
    try {
      instance().sendPasswordResetEmail(email: email);
      result = FireResult(success: true, message: "Reset password email has been sent", data: null);
    } on FirebaseAuthException catch (e) {
        result = FireResult(success: false, message: e.toString(), data: null);
    } catch(e){
      result = FireResult(success: false, message: e.toString(), data: null);
    }
    return result;
  }

  Future<FireResult<FireUser>> sendEmailVerification() async{
    FireResult<FireUser> result;
    FireUser user = currentUser();
    if(user != null){
      if(!user.emailVerified()){
        await user.detail.sendEmailVerification();
        result = FireResult<FireUser>(success: true, message: "Email verification has been sent", data: user);
      }
      else{
        result = FireResult<FireUser>(success: true, message: "Email has already been verified", data: user);
      }
    }
    else{
      result = FireResult<FireUser>(success: false, message: "No logged in user found", data: null);
    }
    return result;
  }

  Future<FireResult<FireUser>> verifyEmailCode(String code) async{
    FireResult<FireUser> result;
    FireUser user = currentUser();
    if(user != null){
      if(!user.emailVerified()){
        try {
          await instance().checkActionCode(code);
          await instance().applyActionCode(code);

          // If successful, reload the user:
          instance().currentUser.reload();
          result = FireResult<FireUser>(success: true, message: "Email verification successful", data: currentUser());
        } on FirebaseAuthException catch (e) {
          if (e.code == 'invalid-action-code') {
            print('The code is invalid.');
            result = FireResult<FireUser>(success: false, message: "The code entered is invalid", data: user);
          } else {
            result = FireResult(success: false, message: e.toString(), data: null);
          }
        } catch(e) {
          result = FireResult(success: false, message: e.toString(), data: null);
        }
      }
      else{
        result = FireResult<FireUser>(success: true, message: "Email has already been verified", data: user);
      }
    }
    else{
      result = FireResult<FireUser>(success: false, message: "No logged in user found", data: null);
    }
    return result;
  }

  Future<FireResult<FireUser>> signOut() async{
    FireResult<FireUser> result;
    if(signedIn()){
      FireUser user = currentUser();
      await instance().signOut();
      result = FireResult<FireUser>(success: true, message: "You have successfully signed out", data: user);
    }
    else{
      result = FireResult<FireUser>(success: false, message: "No logged in user found", data: null);
    }
    return result;
  }

  Future<FireResult<FireUser>> deleteAccount() async{
    FireResult<FireUser> result;
    if(signedIn()){
      FireUser user = currentUser();
      try {
        await user.detail.delete();
        result = FireResult<FireUser>(success: true, message: "Your account has been deleted", data: user);
      } on FirebaseAuthException catch (e) {
        if (e.code == 'requires-recent-login') {
          result = FireResult<FireUser>(success: false, message: "You must login again to execute this operation", data: user);
        } else{
          result = FireResult<FireUser>(success: false, message: e.toString(), data: null);
        }
      } catch (e){
        result = FireResult<FireUser>(success: false, message: e.toString(), data: null);
      }
    }
    else{
      result = FireResult<FireUser>(success: false, message: "No logged in user found", data: null);
    }
    return result;
  }

  Future<FireResult<UserCredential>> reauthenticate({String email, String password}) async{
    FireResult<UserCredential> result;
    if(signedIn()) {
      EmailAuthCredential credential = EmailAuthProvider.credential(email: email, password: password);
      try{
        UserCredential userCredential = await instance().currentUser.reauthenticateWithCredential(credential);
        result = FireResult(success: true, message: "Reauthenticate successful", data: userCredential);
      } on FirebaseAuthException catch (e) {
        if (e.code == 'wrong-password') {
          result = FireResult(success: false, message: "Incorrect password was entered", data: null);
        } else {
          result = FireResult(success: false, message: e.toString(), data: null);
        }
      }
      catch(e){
        result = FireResult(success: false, message: e.toString(), data: null);
      }
    }
    else{
      result = FireResult(success: false, message: "No logged in user found", data: null);
    }
    return result;
  }
}

class NUIFireAuthBuilder implements NUIFireBuilder<NUIFireAuth>{
  FireAuthStateChangeListener _authStateChangeListener;

  NUIFireAuthBuilder setAuthStateChangeListener(FireAuthStateChangeListener listener){
    this._authStateChangeListener = listener;
    return this;
  }

  @override
  NUIFireAuth build(FirebaseApp app) {
    NUIFireAuth auth = NUIFireAuth._builder(this);
    auth.listen(_authStateChangeListener);
    NUIFireAuth._authInstance = auth;
    return auth;
  }
}