import 'dart:async';
import 'dart:developer';
import 'package:firebase_core/firebase_core.dart';
import 'package:flutter/material.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_firebase/nui_firebase.dart';
import 'package:nui_firebase/nui_geohash.dart';

typedef FireCollectionQuery = Query Function(CollectionReference collection);
typedef FireTransactionTask<T> = T Function(Transaction transaction, FireTransaction fireT);
typedef FireBatchTask = void Function(WriteBatch batch, FireBatch fireB);
typedef PreUpdateTask = void Function(Map<String, dynamic> updates);
typedef PreInsertTask = void Function(Map<String, dynamic> data);

class NUIFirestore implements NUIFireComponent{
  NUIFirestore._builder(NUIFirestoreBuilder builder){
    _entities = builder._entities;
    _preInsertTask = builder._preInsertTask;
    _preUpdateTask = builder._preUpdateTask;
    _instance = this;
  }

  Map<String, NUIEntMapper> _entities;
  PreInsertTask _preInsertTask;
  PreUpdateTask _preUpdateTask;
  static NUIFirestore _instance;

  @override
  String name(){
    return "NUIFirestore";
  }

  static NUIFirestore get(){
    if(_instance == null) throw Exception("NUIFirestore -> You have not initialized NUIFirestore with builder");
    return _instance;
  }

  Map<String, NUIEntMapper> getEntities(){
    return _entities;
  }

  FirebaseFirestore _fireInstance(){
    return FirebaseFirestore.instance;
  }

  CollectionReference collection(String collection, {DocumentReference documentReference}){
    if(documentReference == null) return _fireInstance().collection(collection);
    return documentReference.collection(collection);
  }

  DocumentReference documentFromPath(String path){
    return _fireInstance().doc(path);
  }

  Stream<QuerySnapshot> listenToCollection(CollectionReference collection, {bool includeMetadataChanges: false}){
    return collection.snapshots(includeMetadataChanges: includeMetadataChanges ?? false);
  }

  Stream<FireResult<List<T>>> listenToEntityCollection<T extends NUIEnt>({@required CollectionReference collection, NUIEntMapper entityMapper, bool includeMetadataChanges}) {
    NUIEntMapper<T> entity = entityMapper ?? _entities[collection.id];
    throwEntityNotRegistered(entity, collection);

    // ignore: close_sinks
    StreamController<FireResult<List<T>>> controller = StreamController();
    Stream<FireResult<List<T>>> stream = controller.stream;
    listenToCollection(collection, includeMetadataChanges: includeMetadataChanges)
        .listen((snapshot) {
          NUIAsync.run(() {
            if(snapshot.size <= 0){
              final result = FireResult<List<T>>(success: true, message: "No snapshots found", data: List<T>());
              controller.add(result);
            }
            else{
              final result = FireResult<List<T>>(success: true, message: "${snapshot.size} documents found", data: snapshot.docs.map((snapshot) {
                return entity.toEntity(snapshot.data());
              }).toList());
              controller.add(result);
            }
            return;
          });
    });
    return stream;
  }

  Stream<DocumentSnapshot> listenToDocument(DocumentReference document, {bool includeMetadataChanges: false}){
    return document.snapshots(includeMetadataChanges: includeMetadataChanges ?? false);
  }

  Stream<FireResult<T>> listenToEntityDocument<T extends NUIEnt>({@required DocumentReference document, NUIEntMapper entityMapper, bool includeMetadataChanges}) {
    NUIEntMapper<T> entity = entityMapper ?? _entities[document.parent.id];
    throwEntityNotRegistered(entity, document.parent);

    // ignore: close_sinks
    StreamController<FireResult<T>> controller = StreamController();
    Stream<FireResult<T>> stream = controller.stream;
    listenToDocument(document, includeMetadataChanges: includeMetadataChanges)
        .listen((snapshot) {
      NUIAsync.run(() {
        if(snapshot.exists){
          final result = FireResult<T>(success: false, message: "Document does not exists anymore", data: null);
          controller.add(result);
        }
        else{
          final result = FireResult<T>(success: false, message: "Document changed", data: entity.toEntity(snapshot.data()));
          controller.add(result);
        }
        return;
      });
    });
    return stream;
  }

  Future<FireResult> save(DocumentReference document, NUIEnt data) async{
    return saveRaw(document, data.toMap());
  }

  Future<FireResult> saveRaw(DocumentReference document, Map<String, dynamic> data) async{
    try{
      if(_preInsertTask != null) _preInsertTask(data);
      await document.set(data);
      log("NUIFirestore -> Data saved");
      return FireResult(success: true, message: "Data saved successfully");
    }catch(e){
      log("NUIFirestore -> Failed to save data");
      return FireResult(success: false, message: "Failed to save data with error : ${e.toString()}");
    }
  }

  Future<FireResult> update(DocumentReference document, NUIEnt data, {Map<String, double> atomic}) async{
    return updateRaw(document, data.toMap(), atomic: atomic);
  }

  Future<FireResult> updateRaw(DocumentReference document, Map<String, dynamic> data, {Map<String, double> atomic}) async{
    try{
      if(atomic != null){
        for(String key in atomic.keys){
          data[key] = FieldValue.increment(atomic[key]);
        }
      }
      if(_preUpdateTask != null) _preUpdateTask(data);
      await document.update(data);
      log("NUIFirestore -> Data of updated successfully");
      return FireResult(success: true, message: "Data updated successfully");
    }catch(e){
      log("NUIFirestore -> Failed to update data");
      return FireResult(success: false, message: "Failed to update data with error : ${e.toString()}");
    }
  }

  Future<FireResult> delete(DocumentReference document) async{
    try{
      await document.delete();
      log("NUIFirestore -> Document of path ${document.path} has been deleted");
      return FireResult(success: true, message: "Document has been deleted");
    }catch(e){
      log("NUIFirestore -> Failed to delete data of path ${document.path}");
      return FireResult(success: false, message: "Failed to delete document with error : ${e.toString()}");
    }
  }

  Future<FireResult> saveToCollection({@required CollectionReference collection, @required NUIEnt data, String documentId}) async{
    if(isNullOrEmpty(documentId)){
      try{
        final dataMap = data.toMap();
        if(_preInsertTask != null) _preInsertTask(dataMap);
        await collection.add(dataMap);
        log("NUIFirestore -> Data of type ${data.mapper().entityName()} saved to collection ${collection.id}");
        return FireResult(success: true, message: "Data saved successfully to collection");
      }catch(e){
        log("NUIFirestore -> Failed to save data of type ${data.mapper().entityName()} to collection ${collection.id}");
        return FireResult(success: false, message: "Failed to saved data to collection with error : ${e.toString()}");
      }
    }
    else{
      return save(collection.doc(documentId), data);
    }
  }

  Future<FireResult> updateFromCollection({@required CollectionReference collection, @required NUIEnt data, String documentId, Map<String, double> atomic}) async{
    return update(collection.doc(documentId), data, atomic: atomic);
  }

  Future<FireResult> deleteFromCollection({@required CollectionReference collection, String documentId}) async{
    return delete(collection.doc(documentId));
  }

  Future<FireResult<T>> read<T extends NUIEnt>({@required DocumentReference document, NUIEntMapper<T> entityMapper}) async{
    NUIEntMapper<T> entity = entityMapper ?? _entities[document.parent.id];
    throwEntityNotRegistered(entity, document.parent);

    try{
      DocumentSnapshot snapshot = await document.get();
      if(!snapshot.exists) return FireResult<T>(data: null, success: false, message: "Document does not exists");
      return FireResult<T>(data: entity.toEntity(snapshot.data()), success: true, message: "Get document successful");
    }catch(e){
      return FireResult<T>(success: false, message: "Failed to read document with error : ${e.toString()}", data: null);
    }
  }

  Future<FireListResult<T, DocumentSnapshot>> readCollection<T extends NUIEnt>({@required CollectionReference collection, NUIEntMapper<T> entityMapper, FireCollectionQuery query}) async{
    NUIEntMapper<T> entity = entityMapper ?? _entities[collection.id];
    throwEntityNotRegistered(entity, collection);

    try {
      QuerySnapshot snapshot;
      if(query == null){
        snapshot = await collection.get();
      }
      else{
        snapshot = await query(collection).get();
      }

      if (snapshot.size <= 0) {
        log("NUIFirestore -> No documents found in the collection ${collection.path}");
        final result = FireListResult<T, DocumentSnapshot>(success: true, message: "No records found", data: List<T>(), cursor: null);
        return result;
      }
      else {
        log("NUIFirestore -> ${snapshot.size} documents found in the collection ${collection.path}");
        final result = FireListResult<T, DocumentSnapshot>(success: true, message: "${snapshot.size} records found", data: snapshot.docs.map((snapshot) {
          return entity.toEntity(snapshot.data());
        }).toList(), cursor: snapshot.docs.last);
        return result;
      }
    }catch(e){
      log("NUIFirestore -> Failed to read collection with error : ${e.toString()}");
      return FireListResult<T, DocumentSnapshot>(success: false, message: "Failed to read collection with error : ${e.toString()}", data: null, cursor: null);
    }
  }

  Future<FireResult<T>> runTransaction<T>({FireTransactionTask<Future<T>> task}) async{
    try{
      T transactionResult = await _fireInstance().runTransaction((transaction) async {
        return task(transaction, FireTransaction(transaction, get()));
      });
      log("NUIFirestore -> Transaction completed");
      return FireResult<T>(success: true, message: "Transaction run successfully", data: transactionResult);
    }catch(e){
      log("NUIFirestore -> Transaction failed with error : ${e.toString()}");
      return FireResult<T>(success: false, message: "Transaction failed with error : ${e.toString()}", data: null);
    }
  }

  Future<FireResult> runBatch({FireBatchTask task}) async{
    try{
      WriteBatch batch = _fireInstance().batch();
      final FireBatch fireBatch = FireBatch(batch, this);
      task(batch, fireBatch);
      batch.commit();
      return FireResult(success: true, message: "Batch run successfully", data: null);
    }catch(e){
      log("NUIFirestore -> Batch run failed with error : ${e.toString()}");
      return FireResult(success: false, message: "Batch run failed with error : ${e.toString()}", data: null);
    }
  }

  NUIGeoFirestore geo(CollectionReference collection){
    return NUIGeoFirestore(this, collection);
  }

  void throwEntityNotRegistered(NUIEntMapper entity, CollectionReference collection){
    if(entity == null){
      throw Exception("NUIFirestore -> No entity registered for collection ${collection.id}, please register the entity on the builder");
    }
  }
}

class FireBatch{
  FireBatch(this._batch, this._firestore);
  WriteBatch _batch;
  NUIFirestore _firestore;

  void save(DocumentReference document, NUIEnt data){
    return saveRaw(document, data.toMap());
  }

  void saveRaw(DocumentReference document, Map<String, dynamic> data){
    if(_firestore._preInsertTask != null) _firestore._preInsertTask(data);
    return _batch.set(document, data);
  }

  void update(DocumentReference document, NUIEnt data, {Map<String, double> atomic}){
    return updateRaw(document, data.toMap(), atomic: atomic);
  }

  void updateRaw(DocumentReference document, Map<String, dynamic> data, {Map<String, double> atomic}){
    if(atomic != null){
      for(String key in atomic.keys){
        data[key] = FieldValue.increment(atomic[key]);
      }
    }
    if(_firestore._preUpdateTask != null) _firestore._preUpdateTask(data);
    return _batch.update(document, data);
  }

  void delete(DocumentReference document){
    return _batch.delete(document);
  }
}

class FireTransaction{
  FireTransaction(this._transaction, this._firestore);
  final Transaction _transaction;
  final NUIFirestore _firestore;

  Transaction save(DocumentReference document, NUIEnt data){
    return saveRaw(document, data.toMap());
  }

  Transaction saveRaw(DocumentReference document, Map<String, dynamic> data){
    if(_firestore._preInsertTask != null) _firestore._preInsertTask(data);
    return _transaction.set(document, data);
  }

  Transaction update(DocumentReference document, NUIEnt data, {Map<String, double> atomic}){
    return updateRaw(document, data.toMap(), atomic: atomic);
  }

  Transaction updateRaw(DocumentReference document, Map<String, dynamic> data, {Map<String, double> atomic}){
    if(atomic != null){
      for(String key in atomic.keys){
        data[key] = FieldValue.increment(atomic[key]);
      }
    }
    if(_firestore._preUpdateTask != null) _firestore._preUpdateTask(data);
    return _transaction.update(document, data);
  }

  Transaction delete(DocumentReference document){
    return _transaction.delete(document);
  }

  Future<T> read<T extends NUIEnt>({@required DocumentReference document, NUIEntMapper<T> entityMapper}) async{
    NUIEntMapper entity = entityMapper ?? _firestore._entities[document.parent.id];
    _firestore.throwEntityNotRegistered(entity, document.parent);

    DocumentSnapshot snapshot = await _transaction.get(document);
    if(!snapshot.exists) return null;
    return entity.toEntity(snapshot.data());
  }
}

class NUIFirestoreBuilder implements NUIFireBuilder<NUIFirestore>{
  Map<String, NUIEntMapper> _entities;
  PreUpdateTask _preUpdateTask;
  PreInsertTask _preInsertTask;

  NUIFirestoreBuilder setEntities(Map<String, NUIEntMapper> entities){
    this._entities = entities;
    return this;
  }

  NUIFirestoreBuilder preUpdateTask(PreUpdateTask task){
    this._preUpdateTask = task;
    return this;
  }

  NUIFirestoreBuilder preInsertTask(PreInsertTask task){
    this._preInsertTask = task;
    return this;
  }

  @override
  NUIFirestore build(FirebaseApp app){
    NUIFirestore firestore = NUIFirestore._builder(this);
    return firestore;
  }
}