import 'dart:math';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/cupertino.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_database/models/nui_database_entity_field.dart';
import 'package:nui_firebase/nui_firebase.dart';
import 'nui_firebase_models.dart';

const E2 = 0.00669447819799;
const EARTH_EQ_RADIUS = 6378137.0;
const EARTH_MERIDIONAL_CIRCUMFERENCE = 40007860.0;
const EPSILON = 1e-12;
const MAX_SUPPORTED_RADIUS = 8587;
const METERS_PER_DEGREE_LATITUDE = 110574.0;

class WithGeoHash {
  static List<NUIDBField> dbFields<T extends WithGeoHash>() {
    return [
      NUIDBField<T>(name: "g", type: NUIEntType.STRING, setter: (data, value) => data.g = value, getter: (data) => data.g),
      NUIDBField<T>(name: "l", type: NUIEntType.LIST_DOUBLE, setter: (data, value) => data.l = asList<double>(value), getter: (data) => data.l),
      NUIDBField<T>(name: "d", type: NUIEntType.DOUBLE, setter: (data, value) => data.d = value, getter: (data) => data.d, ignore: true),
    ];
  }
  String g;
  List<double> l;
  double d;
}

///
/// A GeoFirestore instance is used to store and query geo location data in Firestore.
///
class NUIGeoFirestore {
  NUIFirestore _firestore;
  CollectionReference collectionReference;

  NUIGeoFirestore(this._firestore, this.collectionReference);

  ///
  /// Build a GeoPoint from a [documentSnapshot]
  ///
  static GeoPoint getLocationValue(DocumentSnapshot documentSnapshot) {
    try {
      final data = documentSnapshot.data();
      if (data != null && data['l'] != null) {
        List<double> latlng = data["l"];
        final GeoPoint location = GeoPoint(latlng[0], latlng[1]);
        final latitude = location.latitude;
        final longitude = location.longitude;
        if (NUIGeoUtils.coordinatesValid(latitude, longitude)) {
          return location;
        }
      }
      return null;
    } catch (e) {
      print('Error occurred when getLocationValue: ' + e.toString());
      return null;
    }
  }

  static Map<String, dynamic> attachGeoHash(Map<String, dynamic> data, GeoPoint location){
    var geoHash = NUIGeoHash.encode(location.latitude, location.longitude);
    data['g'] = geoHash;
    data['l'] = [location.latitude, location.longitude];
    return data;
  }

  static WithGeoHash attachGeoHashToItem(WithGeoHash data, GeoPoint location){
    var geoHash = NUIGeoHash.encode(location.latitude, location.longitude);
    data.g = geoHash;
    data.l = [location.latitude, location.longitude];
    return data;
  }

  /// Sets the [location] of a document for the given [documentID].
  Future<dynamic> setLocation(String documentID, GeoPoint location) async {
    if (documentID == null) {
      throw FormatException('Document ID is null');
    }
    var docRef = this.collectionReference.doc(documentID);
    var geoHash = NUIGeoHash.encode(location.latitude, location.longitude);
    // Create a Map with the fields to add
    var updates = Map<String, dynamic>();
    updates['g'] = geoHash;
    updates['l'] = [location.latitude, location.longitude];
    // Update the DocumentReference with the location data
    return await docRef.update(updates);
  }

  ///
  /// Removes the [location] of a document for the given [documentID].
  ///
  Future<dynamic> removeLocation(String documentID, GeoPoint location) async {
    if (documentID == null) {
      throw FormatException('Document ID is null');
    }
    //Get the DocumentReference for this documentID
    var docRef = this.collectionReference.doc(documentID);
    //Create a Map with the fields to add
    var updates = Map<String, dynamic>();
    updates['g'] = null;
    updates['l'] = null;
    //Update the DocumentReference with the location data
    await docRef.update(updates);
  }

  ///
  /// Gets the current location of a document for the given [documentID].
  ///
  Future<GeoPoint> getLocation(String documentID) async {
    final snapshot = await this.collectionReference.doc(documentID).get();
    final geoPoint = getLocationValue(snapshot);
    return geoPoint;
  }

  ///
  /// Returns the documents centered at a given location and with the given radius.
  /// [center]      The center of the query
  /// [radius]      The radius of the query, in kilometers. The maximum radius that is
  ///               supported is about 8587km. If a radius bigger than this is passed we'll cap it.
  /// [addDistance] Whether to process data and add distance property to returned documents, defaults to True.
  /// [exact]       Whether to process data and remove documents that are further than specified radius, defaults to True.
  ///
  Future<FireResult<List<T>>> getAtLocation<T extends NUIEnt>({@required GeoPoint center, @required double radiusKm, @required NUIEntMapper entityMapper, bool exact = true, bool addDistance = true}) async {
    NUIEntMapper<T> entity = entityMapper ?? _firestore.getEntities()[collectionReference.id];
    _firestore.throwEntityNotRegistered(entity, collectionReference);

    //Get the futures from Firebase Queries generated from GeoHashQueries
    final futures = NUIGeoHashQuery.queriesAtLocation(
        center, NUIGeoUtils.capRadius(radiusKm) * 1000)
        .map((query) => query.createFirestoreQuery(this).get());

    //Await the completion of all the futures
    try {
      List<T> documents = [];
      final snapshots = await Future.wait(futures);
      snapshots.forEach((snapshot) {
        snapshot.docs.forEach((doc) {
          if (addDistance || exact) {
            final List<dynamic> latlng = doc.data()['l'];
            final distance = !isNullOrEmpty(latlng) ? NUIGeoUtils.distance(center, GeoPoint(latlng[0], latlng[1])) : null;
            print("Distance : $distance");
            if(distance != null) {
              if (exact) {
                if (distance <= radiusKm) {
                  doc.data()['d'] = distance;
                  documents.add(entity.toEntity(doc.data()));
                }
              } else {
                doc.data()['d'] = distance;
                documents.add(entity.toEntity(doc.data()));
              }
            }
            else{
              documents.add(entity.toEntity(doc.data()));
            }
          } else {
            documents.add(entity.toEntity(doc.data()));
          }
        });
      });
      if(documents.isEmpty){
        return FireResult<List<T>>(success: true, message: "No data found at the queried location", data: documents);
      }
      else{
        return FireResult<List<T>>(success: true, message: "${documents.length} records found at this location", data: documents);
      }
    } catch (e) {
      print('NUIGeoHash -> Failed retrieving data for geo query: ' + e.toString());
      return FireResult<List<T>>(success: false, message: "Failed to get geo query with error : ${e.toString()}", data: null);
    }
  }
}

class NUIGeoHash{
  // The default precision of a geohash
  static const DEFAULT_PRECISION = 10;

  // The maximal precision of a geohash
  static const MAX_PRECISION = 22;

  // The maximal number of bits precision for a geohash
  static const MAX_PRECISION_BITS =
      MAX_PRECISION * NUIGeoBase32Utils.BITS_PER_BASE32_CHAR;

  /// Encode a latitude and longitude pair into a geohash string.
  static String encode(final double latitude, final double longitude,
      {final int precision: DEFAULT_PRECISION}) {
    if (precision > MAX_PRECISION) {
      throw new ArgumentError('latitude and longitude are not precise enough to encode $precision characters');
    }
    final latitudeBase2 = (latitude + 90) * (pow(2.0, 52) / 180);
    final longitudeBase2 = (longitude + 180) * (pow(2.0, 52) / 360);
    final longitudeBits = (precision ~/ 2) * 5 + (precision % 2) * 3;
    final latitudeBits = precision * 5 - longitudeBits;
    var longitudeCode = longitudeBase2.floor() >> (52 - longitudeBits);
    var latitudeCode = latitudeBase2.floor() >> (52 - latitudeBits);

    final stringBuffer = [];
    for (var localPrecision = precision; localPrecision > 0; localPrecision--) {
      int bigEndCode, littleEndCode;
      if (localPrecision % 2 == 0) {
        // Even slot. Latitude is more significant.
        bigEndCode = latitudeCode;
        littleEndCode = longitudeCode;
        latitudeCode >>= 3;
        longitudeCode >>= 2;
      } else {
        bigEndCode = longitudeCode;
        littleEndCode = latitudeCode;
        latitudeCode >>= 2;
        longitudeCode >>= 3;
      }
      final code = ((bigEndCode & 4) << 2) |
      ((bigEndCode & 2) << 1) |
      (bigEndCode & 1) |
      ((littleEndCode & 2) << 2) |
      ((littleEndCode & 1) << 1);
      stringBuffer.add(NUIGeoBase32Utils.valueToBase32Char(code));
    }
    final buffer = new StringBuffer()..writeAll(stringBuffer.reversed);
    return buffer.toString();
  }

  /// Get the rectangle that covers the entire area of a geohash string.
  static Rectangle<double> getExtents(String geohash) {
    final precision = geohash.length;
    if (precision > MAX_PRECISION) {
      throw new ArgumentError(
          'latitude and longitude are not precise enough to encode $precision characters');
    }
    var latitudeInt = 0;
    var longitudeInt = 0;
    var longitudeFirst = true;
    for (var character
    in geohash.codeUnits.map((r) => new String.fromCharCode(r))) {
      int thisSequence;
      try {
        thisSequence = NUIGeoBase32Utils.base32CharToValue(character);
      } catch (error) {
        throw new ArgumentError('$geohash was not a geohash string');
      }
      final bigBits = ((thisSequence & 16) >> 2) |
      ((thisSequence & 4) >> 1) |
      (thisSequence & 1);
      final smallBits = ((thisSequence & 8) >> 2) | ((thisSequence & 2) >> 1);
      if (longitudeFirst) {
        longitudeInt = (longitudeInt << 3) | bigBits;
        latitudeInt = (latitudeInt << 2) | smallBits;
      } else {
        longitudeInt = (longitudeInt << 2) | smallBits;
        latitudeInt = (latitudeInt << 3) | bigBits;
      }
      longitudeFirst = !longitudeFirst;
    }
    final longitudeBits = (precision ~/ 2) * 5 + (precision % 2) * 3;
    final latitudeBits = precision * 5 - longitudeBits;

    longitudeInt = longitudeInt << (52 - longitudeBits);
    latitudeInt = latitudeInt << (52 - latitudeBits);
    final longitudeDiff = 1 << (52 - longitudeBits);
    final latitudeDiff = 1 << (52 - latitudeBits);
    final latitude = latitudeInt.toDouble() * (180 / pow(2.0, 52)) - 90;
    final longitude = longitudeInt.toDouble() * (360 / pow(2.0, 52)) - 180;
    final height = latitudeDiff.toDouble() * (180 / pow(2.0, 52));
    final width = longitudeDiff.toDouble() * (360 / pow(2.0, 52));
    return Rectangle<double>(latitude + height, longitude, height, width);
    //I know this is backward, but it's because lat/lng are backwards.
  }

  /// Get a single number that is the center of a specific geohash rectangle.
  static Point<double> decode(String geohash) {
    final extents = getExtents(geohash);
    final x = extents.left + extents.width / 2;
    final y = extents.bottom + extents.height / 2;
    return new Point<double>(x, y);
  }
}

class NUIGeoHashQuery {
  final String startValue;
  final String endValue;

  NUIGeoHashQuery({this.startValue, this.endValue});

  static double bitsLatitudeForResolution(double resolution) {
    return min(log(EARTH_MERIDIONAL_CIRCUMFERENCE / 2 / resolution) / log(2.0),
        NUIGeoHash.MAX_PRECISION_BITS.toDouble());
  }

  static double bitsLongitudeForResolution(double resolution, double latitude) {
    final degrees = NUIGeoUtils.distanceToLongitudeDegrees(resolution, latitude);
    return degrees.abs() > 0 ? max(1.0, log(360 / degrees) / log(2.0)) : 1.0;
  }

  static int bitsForBoundingBox(GeoPoint location, double size) {
    final latitudeDegreesDelta = NUIGeoUtils.distanceToLatitudeDegrees(size);
    final latitudeNorth = min(90.0, location.latitude + latitudeDegreesDelta);
    final latitudeSouth = max(-90.0, location.latitude - latitudeDegreesDelta);
    final bitsLatitude = (bitsLatitudeForResolution(size).floor() * 2).toInt();
    final bitsLongitudeNorth =
    (bitsLongitudeForResolution(size, latitudeNorth).floor() * 2 - 1)
        .toInt();
    final bitsLongitudeSouth =
    (bitsLongitudeForResolution(size, latitudeSouth).floor() * 2 - 1)
        .toInt();
    return min(bitsLatitude, min(bitsLongitudeNorth, bitsLongitudeSouth));
  }

  static NUIGeoHashQuery queryForGeoHash(String hash, int bits) {
    final precision =
    (bits.toDouble() / NUIGeoBase32Utils.BITS_PER_BASE32_CHAR).ceil().toInt();
    if (hash.length < precision) {
      return NUIGeoHashQuery(startValue: hash, endValue: '$hash~');
    }
    hash = hash.substring(0, precision);
    final base = hash.substring(0, hash.length - 1);
    final lastValue = NUIGeoBase32Utils.base32CharToValue(hash[hash.length - 1]);
    final significantBits =
        bits - (base.length * NUIGeoBase32Utils.BITS_PER_BASE32_CHAR);
    final unusedBits = NUIGeoBase32Utils.BITS_PER_BASE32_CHAR - significantBits;
    // delete unused bits
    final startValue = (lastValue >> unusedBits) << unusedBits;
    final endValue = startValue + (1 << unusedBits);
    final startHash = base + NUIGeoBase32Utils.valueToBase32Char(startValue);
    final endHash = (endValue > 31)
        ? '$base~'
        : base + NUIGeoBase32Utils.valueToBase32Char(endValue);
    return NUIGeoHashQuery(startValue: startHash, endValue: endHash);
  }

  static Set<NUIGeoHashQuery> queriesAtLocation(GeoPoint location, double radius) {
    final queryBits = max(1, bitsForBoundingBox(location, radius));
    final geoHashPrecision =
    (queryBits.toDouble() / NUIGeoBase32Utils.BITS_PER_BASE32_CHAR)
        .ceil()
        .toInt();

    final latitude = location.latitude;
    final longitude = location.longitude;
    final latitudeDegrees = radius / METERS_PER_DEGREE_LATITUDE;
    final latitudeNorth = min(90.0, latitude + latitudeDegrees);
    final latitudeSouth = max(-90.0, latitude - latitudeDegrees);
    final longitudeDeltaNorth = NUIGeoUtils.distanceToLongitudeDegrees(radius, latitudeNorth);
    final longitudeDeltaSouth = NUIGeoUtils.distanceToLongitudeDegrees(radius, latitudeSouth);
    final longitudeDelta = max(longitudeDeltaNorth, longitudeDeltaSouth);

    final queries = Set<NUIGeoHashQuery>();

    final geoHash =
    NUIGeoHash.encode(latitude, longitude, precision: geoHashPrecision);
    final geoHashW = NUIGeoHash.encode(
        latitude, NUIGeoUtils.wrapLongitude(longitude - longitudeDelta),
        precision: geoHashPrecision);
    final geoHashE = NUIGeoHash.encode(
        latitude, NUIGeoUtils.wrapLongitude(longitude + longitudeDelta),
        precision: geoHashPrecision);
    final geoHashN =
    NUIGeoHash.encode(latitudeNorth, longitude, precision: geoHashPrecision);
    final geoHashNW = NUIGeoHash.encode(
        latitudeNorth, NUIGeoUtils.wrapLongitude(longitude - longitudeDelta),
        precision: geoHashPrecision);
    final geoHashNE = NUIGeoHash.encode(
        latitudeNorth, NUIGeoUtils.wrapLongitude(longitude + longitudeDelta),
        precision: geoHashPrecision);
    final geoHashS =
    NUIGeoHash.encode(latitudeSouth, longitude, precision: geoHashPrecision);
    final geoHashSW = NUIGeoHash.encode(
        latitudeSouth, NUIGeoUtils.wrapLongitude(longitude - longitudeDelta),
        precision: geoHashPrecision);
    final geoHashSE = NUIGeoHash.encode(
        latitudeSouth, NUIGeoUtils.wrapLongitude(longitude + longitudeDelta),
        precision: geoHashPrecision);

    queries.add(queryForGeoHash(geoHash, queryBits));
    queries.add(queryForGeoHash(geoHashE, queryBits));
    queries.add(queryForGeoHash(geoHashW, queryBits));
    queries.add(queryForGeoHash(geoHashN, queryBits));
    queries.add(queryForGeoHash(geoHashNE, queryBits));
    queries.add(queryForGeoHash(geoHashNW, queryBits));
    queries.add(queryForGeoHash(geoHashS, queryBits));
    queries.add(queryForGeoHash(geoHashSE, queryBits));
    queries.add(queryForGeoHash(geoHashSW, queryBits));

    // Join queries
    bool didJoin;
    do {
      NUIGeoHashQuery query1;
      NUIGeoHashQuery query2;
      for (NUIGeoHashQuery query in queries) {
        for (NUIGeoHashQuery other in queries) {
          if (query != other && query._canJoinWith(other)) {
            query1 = query;
            query2 = other;
            break;
          }
        }
      }
      didJoin = (query1 != null && query2 != null);
      if (didJoin) {
        queries.remove(query1);
        queries.remove(query2);
        queries.add(query1._joinWith(query2));
      }
    } while (didJoin);
    return queries;
  }

  bool _isPrefix(NUIGeoHashQuery other) =>
      (other.endValue.compareTo(this.startValue)) >= 0 &&
          (other.startValue.compareTo(this.startValue) < 0) &&
          (other.endValue.compareTo(this.endValue) < 0);

  bool _isSuperQuery(NUIGeoHashQuery other) {
    return other.startValue.compareTo(this.startValue) <= 0 &&
        other.endValue.compareTo(endValue) >= 0;
  }

  bool _canJoinWith(NUIGeoHashQuery other) =>
      this._isPrefix(other) ||
          other._isPrefix(this) ||
          this._isSuperQuery(other) ||
          other._isSuperQuery(this);

  NUIGeoHashQuery _joinWith(NUIGeoHashQuery other) {
    if (other._isPrefix(this)) {
      return NUIGeoHashQuery(
          startValue: this.startValue, endValue: other.endValue);
    }
    if (this._isPrefix(other)) {
      return NUIGeoHashQuery(
          startValue: other.startValue, endValue: this.endValue);
    }
    if (this._isSuperQuery(other)) {
      return other;
    }
    if (other._isSuperQuery(this)) {
      return this;
    }
    throw FormatException("Can't join these two queries: $this, $other");
  }

  Query createFirestoreQuery(NUIGeoFirestore geoFirestore) {
    return geoFirestore.collectionReference
        .orderBy('g')
        .startAt([this.startValue]).endAt([this.endValue]);
  }

  bool containsGeoHash(String hash) {
    return this.startValue.compareTo(hash) <= 0 &&
        this.endValue.compareTo(hash) > 0;
  }

  bool operator ==(dynamic other) {
    if (other == null || !(other is NUIGeoHashQuery)) return false;
    if (endValue != other.endValue || startValue != other.startValue)
      return false;
    return true;
  }

  int get hashCode => (31 * startValue.hashCode + endValue.hashCode);

  toString() => "GeoHashQuery(startValue='$startValue', endValue='$endValue')";
}

class NUIGeoBase32Utils{
  NUIGeoBase32Utils._builder();
  //  number of bits per base 32 character
  static const BITS_PER_BASE32_CHAR = 5;
  //String representing the Base32 character map
  static const _BASE32_CHARS = '0123456789bcdefghjkmnpqrstuvwxyz';

  ///
  /// This method convert a given value to his corresponding Base32 character
  ///
  static String valueToBase32Char(int value) {
    if (value < 0 || value >= _BASE32_CHARS.length)
      throw FormatException("Not a valid base32 value: $value");
    return _BASE32_CHARS[value];
  }

  ///
  /// This method convert a given Base32 character to his corresponding value
  ///
  static int base32CharToValue(String base32Char) {
    final value = _BASE32_CHARS.indexOf(base32Char);
    if (value == -1)
      throw FormatException("Not a valid base32 char: $base32Char");
    return value;
  }

  ///
  /// This method check if a given geo hash is valid
  ///
  static bool isValidBase32String(String string) {
    final RegExp regex = new RegExp('^[$_BASE32_CHARS]*\$');
    return regex.hasMatch(string);
  }
}

class NUIGeoUtils{
  NUIGeoUtils._builder();
  ///
  /// Checks if these coordinates are valid geo coordinates.
  /// [latitude]  The latitude must be in the range [-90, 90]
  /// [longitude] The longitude must be in the range [-180, 180]
  /// returns [true] if these are valid geo coordinates
  ///
  static bool coordinatesValid(double latitude, double longitude) {
    return (latitude >= -90 &&
        latitude <= 90 &&
        longitude >= -180 &&
        longitude <= 180);
  }

  ///
  /// Checks if the coordinates  of a GeopPoint are valid geo coordinates.
  /// [latitude]  The latitude must be in the range [-90, 90]
  /// [longitude] The longitude must be in the range [-180, 180]
  /// returns [true] if these are valid geo coordinates
  ///
  static bool geoPointValid(GeoPoint point) {
    return (point.latitude >= -90 &&
        point.latitude <= 90 &&
        point.longitude >= -180 &&
        point.longitude <= 180);
  }

  ///
  /// Wraps the longitude to [-180,180].
  ///
  /// [longitude] The longitude to wrap.
  /// returns The resulting longitude.
  ///
  static double wrapLongitude(double longitude) {
    if (longitude <= 180 && longitude >= -180) {
      return longitude;
    }
    final adjusted = longitude + 180;
    if (adjusted > 0) {
      return (adjusted % 360) - 180;
    }
    // else
    return 180 - (-adjusted % 360);
  }

  static double degreesToRadians(double degrees) {
    return (degrees * pi) / 180;
  }

  ///
  /// Calculates the number of degrees a given distance is at a given latitude.
  /// [distance] The distance to convert.
  /// [latitude] The latitude at which to calculate.
  /// returns the number of degrees the distance corresponds to.
  static double distanceToLongitudeDegrees(double distance, double latitude) {
    final radians = degreesToRadians(latitude);
    final numerator = cos(radians) * EARTH_EQ_RADIUS * pi / 180;
    final denom = 1 / sqrt(1 - E2 * sin(radians) * sin(radians));
    final deltaDeg = numerator * denom;
    if (deltaDeg < EPSILON) {
      return distance > 0 ? 360.0 : 0.0;
    }
    // else
    return min(360.0, distance / deltaDeg);
  }

  ///
  /// Calculates the distance, in kilometers, between two locations, via the
  /// Haversine formula. Note that this is approximate due to the fact that
  /// the Earth's radius varies between 6356.752 km and 6378.137 km.
  /// [p1] The first location given
  /// [p2] The second location given
  /// return the distance, in kilometers, between the two locations.
  ///
  static double distance(GeoPoint p1, GeoPoint p2) {
    final dlat = degreesToRadians(p2.latitude - p1.latitude);
    final dlon = degreesToRadians(p2.longitude - p1.longitude);
    final lat1 = degreesToRadians(p1.latitude);
    final lat2 = degreesToRadians(p2.latitude);

    final r = 6378.137; // WGS84 major axis
    double c = 2 *
        asin(sqrt(pow(sin(dlat / 2), 2) +
            cos(lat1) * cos(lat2) * pow(sin(dlon / 2), 2)));
    return r * c;
  }

  static double distanceToLatitudeDegrees(double distance) =>
      distance / METERS_PER_DEGREE_LATITUDE;

  static double capRadius(double radius) {
    if (radius > MAX_SUPPORTED_RADIUS) {
      print("The radius is bigger than $MAX_SUPPORTED_RADIUS and hence we'll use that value");
      return MAX_SUPPORTED_RADIUS.toDouble();
    }
    return radius;
  }
}